﻿# *** Default expires time is 600 (sec)
# If you want to change expires time, add to mt-config.cgi like this,
#
# CMSTemplateCachePeriod 600
#
# *** Template Tags
#
# <MTCMSCacheBlock key="CategoryEntries" blog_id="$blog_id" object_ds="category" 
#                            object_id="$category_id" by_user="1" children="1" primary="1" ttl="3600">
#     Some template tags here.
#      ( Clear cache when entry ( page ) is in ( primary ) category ( folder ) (Id:$category_id) changes. )
# </MTCMSCacheBlock>
#
#
# <MTCMSCacheBlock key="EntrySummary" blog_id="$blog_id" object_ds="entry" object_id="$entry_id" children="1">
#     Some template tags here. ( Clear cache when entry ( page ), comment and trackback changes. )
# </MTCMSCacheBlock>
#
#
# <MTCMSCacheBlock key="TagsBlock" blog_id="$blog_id" object_ds="tag" object_id="0" children="1">
#     Some template tags here. ( Clear cache when tag and tagged entry changes. )
# </MTCMSCacheBlock>
#
#
# <MTSetVarBlock name="key">RecentEntries_<$MTCategoryId$></MTSetVarBlock>
# <MTRequestCacheBlock key="$key" blog_id="$blog_id">
#     Some template tags here. ( Clear cache when application request exit. )
# </MTRequestCacheBlock>

package MT::Plugin::CMSCache;
use strict;
use MT;
use MT::Plugin;
use MT::Request;
use MT::Blog;
use MT::Entry;
use MT::Page;
use MT::Placement;
use MT::Category;
use MT::Folder;
use MT::TBPing;
use MT::Comment;
use MT::ObjectTag;
use MT::ObjectAsset;
use CMSCache::CMSCache;

our $VERSION = '0.6';
our $SCHEMA_VERSION = '0.06';

use base qw( MT::Plugin );

@MT::Plugin::CMSCache::ISA = qw( MT::Plugin );

my $plugin = new MT::Plugin::CMSCache ( {
    name => 'CMSCache',
    id   => 'CMSCache',
    key =>  'cmscache',
    description => '<MT_TRANS phrase=\'_PLUGIN_DESCRIPTION\'>',
    author_name => 'Alfasado,Inc.',
    author_link => 'http://alfasado.net/',
    version => $VERSION,
    schema_version => $SCHEMA_VERSION,
#     settings => new MT::PluginSettings( [
#         ['cmscache_ttl', { Default => 600 }],
#     ] ),
    l10n_class => 'CMSCache::L10N',
} );

MT->add_plugin( $plugin );

sub init_registry {
    my $plugin = shift;
    $plugin->registry ( {
        object_types => {
            'cmscache' =>  'CMSCache::CMSCache',
        },
        tags => {
            block => {
                'CMSCacheBlock'     => \&_hndlr_cms_cache,
                'RequestCacheBlock' => \&_hndlr_request_cache,
            },
        },
        callbacks => {
            'MT::Blog::post_save'
                => \&_post_save_blog,
            'cms_post_delete.blog'
                => \&_post_delete_blog,
            'MT::Category::post_save'
                => \&_post_save_category,
            'MT::Folder::post_save'
                => \&_post_save_category,
            'cms_post_delete.category'
                => \&_post_delete_category,
            'cms_post_delete.folder'
                => \&_post_delete_category,
            'MT::Placement::post_save'
                => \&_post_save_placement,
            'MT::Entry::post_save'
                => \&_post_save_entry,
            'cms_post_save.entry'
                => \&_save_entry_after,
            'MT::Page::post_save'
                => \&_post_save_entry,
            'cms_post_save.page'
                => \&_save_entry_after,
            'MT::Comment::post_save'
                => \&_post_save_comment,
            'cms_post_delete.comment'
                => \&_post_delete_comment,
            'MT::TBPing::post_save'
                => \&_post_save_ping,
            'cms_post_delete.ping'
                => \&_post_delete_ping,
            'MT::Tag::post_save'
                => \&_post_save_tag,
            'cms_post_delete.tag'
                => \&_post_delete_tag,
            'MT::Asset::post_save'
                => \&_post_save_asset,
            'cms_post_delete.asset'
                => \&_post_delete_asset,
            'cms_delete_permission_filter.tag'
                => \&_pre_delete_tag,
            'cms_delete_permission_filter.entry'
                => \&_delete_entry,
            'cms_delete_permission_filter.page'
                => \&_delete_entry,
            'cms_post_save.template'
                => \&_post_save_template,
        },
        tasks => {
            'clear_template_cache' => {
                label     => 'Clear template cache',
                frequency => 600,
                code      => \&_clear_template_cache,
            },
        },
   } );
}

sub _clear_template_cache {
    my @co = CMSCache::CMSCache->load( { expires => { '>' => time } } );
    my $clear;
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
        $clear = 1;
    }
    return $clear;
}

sub _post_save_template {
    my ( $cb, $app, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $text = $obj->text;
    my @tags = split ( /<MT/i, $text );
    for my $tag ( @tags ) {
        if ( $text =~ m/:{0,1}CMSCacheBlock\s*.*?key="(.*?)".*?>/sig ) {
            my $terms = { key => $1 };
            if ( $blog_id ) {
                $terms->{ blog_id } = $blog_id;
            }
            my @co = CMSCache::CMSCache->load( $terms );
            for my $cache ( @co ) {
                $cache->remove or die $cache->errstr;
            }
        }
    }
    return 1;
}

sub _post_save_blog {
    my ( $cb, $obj, $original ) = @_;
    return unless defined $original;
    my $blog_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_blog:' . $blog_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_blog' ) unless $rcache;
    $r->cache( 'post_save_blog:' . $blog_id, 1 );
    my $terms = { blog_id => [ 0, $blog_id ] };
    if ( $rcache ) {
        if ( $rcache =~ /blog/ ) {
            $terms->{ children } = 1;
        } else {
            $terms->{ object_ds } = 'blog';
        }
    } else {
        $terms->{ object_ds } = 'blog';
    }
    my @co = CMSCache::CMSCache->load( $terms );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
}

sub _post_delete_blog {
    my ( $cb, $app, $obj, $original ) = @_;
    my $blog_id = $obj->id;
    my @co = CMSCache::CMSCache->load( { blog_id => [ 0, $blog_id ] } );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
    @co = CMSCache::CMSCache->load( { object_ds => 'blog', object_id => $blog_id } );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
}

sub _post_save_category {
    my ( $cb, $obj, $original ) = @_;
    return unless defined $original;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $class = $obj->class;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_category:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_category' ) unless $rcache;
    $r->cache( 'post_save_category:' . $obj_id, 1 );
    my $terms = { object_id => [ 0, $obj->id ], object_ds => $class, blog_id => [ 0, $blog_id ] };
    if ( $rcache ) {
        if ( $rcache =~ /category/ ) {
            $terms->{ children } = 1;
        }
    }
    my @co = CMSCache::CMSCache->load( $terms );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
    for my $parent ( $obj->parent_categories ) {
        next if ( !$parent );
        my $terms = { object_id => [ 0, $parent->id ], children => 1, object_ds => $class, blog_id => [ 0, $blog_id ] };
        my @co = CMSCache::CMSCache->load( $terms );
        for my $cache ( @co ) {
            $cache->remove or die $cache->errstr;
        }
    }
}

sub _post_delete_category {
    my ( $cb, $app, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $class = $obj->class;
    my @co = CMSCache::CMSCache->load( { object_id => [ 0, $obj->id ], object_ds => $class, blog_id => [ 0, $blog_id ] } );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
    for my $parent ( $obj->parent_categories ) {
        next if ( !$parent );
        my $terms = { object_id => [ 0, $parent->id ], children => 1, object_ds => $class, blog_id => [ 0, $blog_id ] };
        my @co = CMSCache::CMSCache->load( $terms );
        for my $cache ( @co ) {
            $cache->remove or die $cache->errstr;
        }
    }
}

sub _post_save_placement {
    my ( $cb, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $category_id = $obj->category_id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_placement:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_placement' ) unless $rcache;
    $r->cache( 'post_save_placement:' . $obj_id, 1 );
    my $c = MT::Category->load( $category_id );
    my $class = $c->class;
    my $terms = { object_id => $category_id, object_ds => $class, blog_id => [ 0, $blog_id ] };
    if ( $rcache ) {
        if ( $rcache =~ /category/ ) {
            $terms->{ children } = 1;
        }
    }
    my @co = CMSCache::CMSCache->load( $terms );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
}

sub _post_save_entry {
    my ( $cb, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $class = $obj->class;
    my $parent = 'category';
    $parent = 'folder' if $class eq 'entry';
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_entry:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_entry' ) unless $rcache;
    $r->cache( 'post_save_entry:' . $obj_id, 1 );
    my $terms = { object_id => [ 0, $obj->id ], object_ds => $class, blog_id => [ 0, $blog_id ] };
    if ( $rcache ) {
        if ( $rcache =~ /entry/ ) {
            $terms->{ children } = 1;
        }
    }
    my $categories = $obj->categories;
    my $primary = $obj->category;
    my $primary_id;
    $primary_id = $primary->id if defined $primary;
    for my $category ( @$categories ) {
        my $is_primary;
        if ( $primary_id == $category->id ) {
            $is_primary = 1;
        }
        my $terms = { object_id => $category->id, object_ds => $parent, blog_id => [ 0, $blog_id ], children => 1 };
        my @co = CMSCache::CMSCache->load( $terms );
        for my $cache ( @co ) {
            if ( ( ( $cache->primary ) && ( $is_primary ) ) || (! $cache->primary ) ) {
                $cache->remove or die $cache->errstr;
            }
        }
    }
    return unless defined $original;
    my @co = CMSCache::CMSCache->load( $terms );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
    my $tags = $obj->get_tag_objects;
    my @tids;
    for my $tag ( @$tags ) {
        push ( @tids, $tag->id );
    }
    my $t = join( ',', @tids );
    $r->cache( 'entry_tags:' . $obj_id, $t );
}

sub _save_entry_after {
    my ( $cb, $app, $obj, $original ) = @_;
    # Check for Tag
    my $blog_id = 0;
    if ( ref $app eq 'MT::App::CMS' ) {
        if ( defined $app->blog ) {
            $blog_id = $app->blog->id;
        }
    }
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $orig_t = $r->cache( 'entry_tags:' . $obj_id );
    my $tags = $obj->get_tag_objects;
    my @orig_tids = split( /,/,  $orig_t ) if $orig_t;
    my @tids;
    my @new_tids = ();
    for my $tag ( @$tags ) {
        my $tag_id = $tag->id;
        push ( @tids, $tag_id );
        if ( $orig_t ) {
            if (! grep ( /^$tag_id$/, @orig_tids ) ) {
                push ( @new_tids, $tag_id );
            }
        }
    }
    my $after_tags = join( ',', @tids );
    push ( @tids, @orig_tids );
    my %count;
    @tids = grep { !$count{$_}++ } @tids;
    my $change;
    for my $id ( @tids ) {
        if (! grep ( /^$id$/, @new_tids ) ) {
            $change = 1;
            my $terms = { object_id => $id, object_ds => 'tag', blog_id => [ 0, $blog_id ] };
            if ( ( grep ( /^$id$/, @orig_tids ) ) && ( grep ( /^$id$/, @$tags ) ) ) {
                $terms->{ children } = 1;
            }
            my @co = CMSCache::CMSCache->load( $terms );
            for my $cache ( @co ) {
                $cache->remove or die $cache->errstr;
            }
        }
    }
    if ( ( $change ) || ( @new_tids ) ) {
        my $terms = { object_id => 0, object_ds => 'tag', blog_id => [ 0, $blog_id ] };
        if ( $after_tags eq $orig_t ) {
            $terms->{ children } = 1;
        }
        my @co = CMSCache::CMSCache->load( $terms );
        for my $cache ( @co ) {
            $cache->remove or die $cache->errstr;
        }
    }
}

sub _delete_entry {
    my ( $cb, $app, $obj ) = @_;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $class = $obj->class;
    my $parent = 'category';
    $parent = 'folder' if $class eq 'entry';
    my $r = MT::Request->instance;
    my $self = $r->cache( 'delete_entry:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'delete_entry' ) unless $rcache;
    $r->cache( 'delete_entry:' . $obj_id, 1 );
    my @co = CMSCache::CMSCache->load( { object_id => [ 0, $obj->id ], object_ds => $class, blog_id => [ 0, $blog_id ] } );
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
    my $categories = $obj->categories;
    my $primary = $obj->category;
    my $primary_id;
    $primary_id = $primary->id if defined $primary;
    for my $category ( @$categories ) {
        my $is_primary;
        if ( $primary_id == $category->id ) {
            $is_primary = 1;
        }
        my $terms = { object_id => $category->id, object_ds => $parent, children => 1 };
        my @co = CMSCache::CMSCache->load( $terms );
        for my $cache ( @co ) {
            if ( ( ( $cache->primary ) && ( $is_primary ) ) || (! $cache->primary ) ) {
                $cache->remove or die $cache->errstr;
            }
        }
    }
    return 1;
}

sub _post_save_comment {
    my ( $cb, $obj, $original ) = @_;
    return unless defined $original;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_comment:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_comment' ) unless $rcache;
    $r->cache( 'post_save_comment:' . $obj_id, 1 );
}

sub _post_delete_comment {
    my ( $cb, $app, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_delete_comment:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'delete_comment' ) unless $rcache;
    $r->cache( 'post_delete_comment:' . $obj_id, 1 );
}

sub _post_save_ping {
    my ( $cb, $obj, $original ) = @_;
    return unless defined $original;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_ping:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_ping' ) unless $rcache;
    $r->cache( 'post_save_ping:' . $obj_id, 1 );
}

sub _post_delete_ping {
    my ( $cb, $app, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_delete_ping:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'delete_ping' ) unless $rcache;
    $r->cache( 'post_delete_ping:' . $obj_id, 1 );
}

sub _post_save_asset {
    my ( $cb, $obj, $original ) = @_;
    my $app = MT->instance;
    return unless defined $original;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_asset:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    my @asset_objects = MT::ObjectAsset->load( { asset_id => $obj_id } );
    for my $asset_obj ( @asset_objects ) {
        my $ds = $asset_obj->object_ds;
        my $object_id = $asset_obj->object_id;
        if ( $ds eq 'entry' ) {
            my $e = MT::Entry->load( $object_id );
            if ( defind $e ) {
                my $class = $e->class;
                my $terms = { object_id => $object_id, object_ds => $class, children => 1 };
                my @co = CMSCache::CMSCache->load( $terms );
                for my $cache ( @co ) {
                    $cache->remove or die $cache->errstr;
                }
            }
        }
    }
    $r->cache( 'first_req:' . $blog_id, 'save_asset' ) unless $rcache;
    $r->cache( 'post_save_asset:' . $obj_id, 1 );
}

sub _post_delete_asset {
    my ( $cb, $app, $obj, $original ) = @_;
    my $blog_id = $obj->blog_id;
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_delete_asset:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'delete_asset' ) unless $rcache;
    my @asset_objects = MT::ObjectAsset->load( { asset_id => $obj_id } );
    $r->cache( 'post_delete_asset:' . $obj_id, 1 );
    for my $asset_obj ( @asset_objects ) {
        my $ds = $asset_obj->object_ds;
        my $object_id = $asset_obj->object_id;
        if ( $ds eq 'entry' ) {
            my $e = MT::Entry->load( $object_id );
            if ( defind $e ) {
                my $class = $e->class;
                my $terms = { object_id => $object_id, object_ds => $class, children => 1 };
                my @co = CMSCache::CMSCache->load( $terms );
                for my $cache ( @co ) {
                    $cache->remove or die $cache->errstr;
                }
            }
        }
    }
}

sub _post_save_tag {
    my ( $cb, $obj, $original ) = @_;
    my $blog_id = 0;
    my $app = MT->instance;
    if ( ref $app eq 'MT::App::CMS' ) {
        if ( defined $app->blog ) {
            $blog_id = $app->blog->id;
        }
    }
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_save_tag:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'save_tag' ) unless $rcache;
    $r->cache( 'post_save_tag:' . $obj_id, 1 );
    if ( (! $rcache ) || ( $rcache !~ /entry/ ) ) {
        if ( defined $original ) {
            my @tagged_objects = MT::ObjectTag->load( { tag_id => $obj_id } );
            for my $tagged_obj ( @tagged_objects ) {
                my $ds = $tagged_obj->object_datasource;
                my $object_id = $tagged_obj->object_id;
                if ( $ds eq 'entry' ) {
                    my $e = MT::Entry->load( $object_id );
                    if ( defind $e ) {
                        my $class = $e->class;
                        my $terms = { object_id => $object_id, object_ds => $class, children => 1 };
                        my @co = CMSCache::CMSCache->load( $terms );
                        for my $cache ( @co ) {
                            $cache->remove or die $cache->errstr;
                        }
                    }
                }
            }
        }
    }
    return unless defined $original;
    my @co;
    if ( $blog_id ) {
        my $terms = { blog_id => [ 0, $blog_id ], object_ds => 'tag', object_id => [ 0, $obj_id ] };
        @co = CMSCache::CMSCache->load( $terms );
    } else {
        my $terms = { object_ds => 'tag', object_id => [ 0, $obj_id ] };
        @co = CMSCache::CMSCache->load( $terms );
    }
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
}

sub _pre_delete_tag {
    my ( $cb, $app, $obj ) = @_;
    my $blog_id = 0;
    if ( ref $app eq 'MT::App::CMS' ) {
        if ( defined $app->blog ) {
            $blog_id = $app->blog->id;
        }
    }
    my $obj_id = $obj->id;
    my $r = MT::Request->instance;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    if ( (! $rcache ) || ( $rcache !~ /entry/ ) ) {
        my @tagged_objects = MT::ObjectTag->load( { tag_id => $obj_id } );
        for my $tagged_obj ( @tagged_objects ) {
            my $ds = $tagged_obj->object_datasource;
            my $object_id = $tagged_obj->object_id;
            if ( $ds eq 'entry' ) {
                my $e = MT::Entry->load( $object_id );
                if ( defind $e ) {
                    my $class = $e->class;
                    my $terms = { object_id => $object_id, object_ds => $class, children => 1 };
                    my @co = CMSCache::CMSCache->load( $terms );
                    for my $cache ( @co ) {
                        $cache->remove or die $cache->errstr;
                    }
                }
            }
        }
    }
}

sub _post_delete_tag {
    my ( $cb, $app, $obj, $original ) = @_;
    my $obj_id = $obj->id;
    my $blog_id = 0;
    if ( ref $app eq 'MT::App::CMS' ) {
        if ( defined $app->blog ) {
            $blog_id = $app->blog->id;
        }
    }
    my $r = MT::Request->instance;
    my $self = $r->cache( 'post_delete_tag:' . $obj_id );
    return if $self;
    my $rcache = $r->cache( 'first_req:' . $blog_id );
    $r->cache( 'first_req:' . $blog_id, 'delete_tag' ) unless $rcache;
    $r->cache( 'post_delete_tag:' . $obj_id, 1 );
    my @co;
    if ( $blog_id ) {
        my $terms = { blog_id => [ 0, $blog_id ], object_ds => 'tag', object_id => [ 0, $obj_id ] };
        @co = CMSCache::CMSCache->load( $terms );
    } else {
        my $terms = { object_ds => 'tag', object_id => [ 0, $obj_id ] };
        @co = CMSCache::CMSCache->load( $terms );
    }
    for my $cache ( @co ) {
        $cache->remove or die $cache->errstr;
    }
}

sub _hndlr_cms_cache {
    my ( $ctx, $args, $cond ) = @_;
    my $key = $args->{ key };
    my $blog_id = $args->{ blog_id };
    my $object_ds = $args->{ object_ds };
    $object_ds = 'blog' unless $object_ds;
    my $children = $args->{ children };
    $children = 1 if $children;
    my $primary = $args->{ primary };
    my $object_id = $args->{ object_id };
    if ( $object_ds eq 'blog' ) {
        if (! $object_id ) {
            $object_id = $blog_id if $blog_id;
        }
    }
    $object_id = 0 unless $object_id;
    $blog_id = 0 unless $blog_id;
    my $ttl = $args->{ ttl };
    my $by_user = $args->{ by_user };
    my $terms = { key => $key };
    $terms->{ blog_id } = $blog_id;
    
    $terms->{ object_ds } = $object_ds if $object_ds;
    $terms->{ object_id } = $object_id;
    $terms->{ children } = $children if $children;
    $terms->{ primary } = $primary if $primary;
    my $app = MT->instance;
    if ( $by_user ) {
        if ( ref $app eq 'MT::App::CMS' ) {
            if ( defined $app->user ) {
                $terms->{ author_id } = $app->user->id;
            }
        }
    }
    my $params = { sort => 'expires', direction => 'descend', limit => 1, };
    my $cache = CMSCache::CMSCache->get_by_key( $terms, $params );
    unless ( $ttl ) {
        # $ttl = $plugin->get_config_value( 'cmscache_ttl' );
        $ttl = $app->config( 'CMSTemplateCachePeriod' ) || 600;
    }
    if ( $cache->id ) {
        if ( time < $cache->expires ) {
            return $cache->data;
        }
    }
    my $tokens = $ctx->stash( 'tokens' );
    my $builder = $ctx->stash( 'builder' );
    my $res = $builder->build( $ctx, $tokens, $cond );
    $cache->data( $res );
    $cache->expires( time + $ttl );
    if ( $object_ds eq 'blog' ) {
        my $r = MT::Request->instance;
        $r->cache( 'post_save_blog:' . $blog_id, 1 );
    }
    $cache->save or die $cache->errstr;
    return $res;
}

sub _hndlr_request_cache {
    my ( $ctx, $args, $cond ) = @_;
    my $key = $args->{ key };
    my $blog_id = $args->{ blog_id };
    $key = "request_cache:$key";
    $key .= ':blog:' . $blog_id if $blog_id;
    my $r = MT::Request->instance;
    my $cache = $r->cache( $key );
    return $cache if $cache;
    my $tokens = $ctx->stash( 'tokens' );
    my $builder = $ctx->stash( 'builder' );
    my $res = $builder->build( $ctx, $tokens, $cond );
    $r->cache( $key, $res );
    return $res;
}

1;